from nonebot.plugin import PluginMetadata

from . import __main__ as __main__
from .config import ConfigModel

__version__ = "0.1.2"
__plugin_meta__ = PluginMetadata(
    name="大电老师活字印刷",
    description="大家好啊，今天来点大家想看的东西",
    usage="直接使用指令【ottohzys】加上文本就行了，不带参数查看详细帮助，不会用我阐释你的梦",
    type="application",
    homepage="https://github.com/lgc-NB2Dev/nonebot-plugin-ottohzys",
    config=ConfigModel,
    supported_adapters={"~onebot.v11"},
    extra={"License": "MIT", "Author": "student_2333"},
)



# Original by: https://github.com/CwavGuy/HUOZI_aolianfeiallin.top/blob/main/huoZiYinShua.py

import asyncio
from dataclasses import dataclass
from io import BytesIO
from pathlib import Path
from typing import List, Tuple, cast

import anyio
import numpy as np
import psola
import soundfile as sf
from async_lru import alru_cache
from nonebot import logger
from numpy.typing import NDArray
from pypinyin import lazy_pinyin

from .resource import CHINGLISH_MAP, TOKENS_DIR, YSDD_TOKEN_MAP, YSDD_TOKENS_DIR

SoundArrayType = NDArray[np.float64]

TARGET_SAMPLE_RATE = 44100
"""目标采样率"""


def normalize_audio(data: SoundArrayType):
    """标准化音频，统一音量"""
    rms = np.sqrt(np.mean(data**2))
    if rms == 0:
        return data
    return data / rms * 0.2


@alru_cache()
async def load_audio(file_path: Path, normalize: bool = True) -> SoundArrayType:
    """读取音频文件"""

    file = BytesIO(await anyio.Path(file_path).read_bytes())
    data, sample_rate = cast(Tuple[SoundArrayType, int], sf.read(file))

    # 双声道转单声道
    if len(data.shape) == 2:
        # 左右声道相加除以2
        data = (data[:, 0] + data[:, 1]) / 2

    # 统一采样率
    if sample_rate != TARGET_SAMPLE_RATE:
        # 计算转换后的长度
        new_length = int((TARGET_SAMPLE_RATE / sample_rate) * len(data))
        # 转换
        data = np.interp(
            np.array(range(new_length)),
            np.linspace(0, new_length - 1, len(data)),
            data,
        )

    # 标准化
    if normalize:
        data = normalize_audio(data)

    return data


@alru_cache()
async def get_pinyin_audio(
    pinyin: str,
    is_ysdd: bool = False,
    normalize: bool = True,
    empty_length: float = 0.25,
) -> SoundArrayType:
    if pinyin != "_":
        file_dir = YSDD_TOKENS_DIR if is_ysdd else TOKENS_DIR
        suffix = ".mp3" if is_ysdd else ".wav"
        file_name = f"{pinyin}{suffix}"
        try:
            return await load_audio(file_dir / file_name, normalize)
        except FileNotFoundError:
            logger.warning(f"Audio {file_name} not found in {file_dir}")

    return np.zeros(int(empty_length * TARGET_SAMPLE_RATE))


def modify_pitch_and_speed(
    data: SoundArrayType,
    pitch_multiple: float,
    speed_multiple: float,
) -> SoundArrayType:
    """改变音高和速度"""

    if pitch_multiple == 1 and speed_multiple == 1:
        # 没有改动的必要，直接返回
        return data

    # 第一次拉伸
    original_len = len(data)
    if speed_multiple / pitch_multiple != 1:
        # 不改变音高的同时在时间上拉伸（PSOLA）
        # constant_stretch 过小会导致 bug，因此分两次拉伸
        data = psola.vocode(
            data,
            TARGET_SAMPLE_RATE,
            constant_stretch=1 / pitch_multiple,
        )
        data = psola.vocode(data, TARGET_SAMPLE_RATE, constant_stretch=speed_multiple)

    # 第二次拉伸，以改变音高的方式拉伸回来
    new_length = int(original_len / speed_multiple)
    return np.interp(
        np.array(range(new_length)),
        np.linspace(0, new_length - 1, len(data)),
        data,
    )


@dataclass
class PreProcPron:
    pron: str
    is_pinyin: bool = False
    is_ysdd: bool = False


async def parse_sentence(sentence: str, ysdd_mode: bool = True) -> List[PreProcPron]:
    pre_proc_pron: List[PreProcPron] = []
    pre_proc_pron.append(PreProcPron(sentence))

    def replace_pron(kw: str, *replace_to: PreProcPron) -> List[PreProcPron]:
        tmp: List[PreProcPron] = []

        for pron in pre_proc_pron:
            if (kw not in pron.pron) or pron.is_pinyin or pron.is_ysdd:
                tmp.append(pron)
            else:
                splitted = [
                    ([*replace_to, PreProcPron(s)] if n else [PreProcPron(s)])
                    for n, s in (enumerate(pron.pron.split(kw)))
                ]
                flattened = [y for x in splitted for y in x if y.pron]
                tmp.extend(flattened)

        return tmp

    # 替换原声大碟
    if ysdd_mode:
        for token, sentences in YSDD_TOKEN_MAP.items():
            for s in sentences:
                pre_proc_pron = replace_pron(s, PreProcPron(token, is_ysdd=True))

    # 替换字母
    for ch, pron in CHINGLISH_MAP.items():
        pre_proc_pron = replace_pron(
            ch,
            *(PreProcPron(pron, is_pinyin=True) for pron in pron.split()),
        )

    # 拼音化
    ret = []
    for pron in pre_proc_pron:
        if pron.is_ysdd or pron.is_pinyin:
            ret.append(pron)
            continue

        for p in lazy_pinyin(pron.pron):
            if p.isalpha():
                ret.append(PreProcPron(p, is_pinyin=True))
            else:
                ret.extend(PreProcPron("_", is_pinyin=True) for _ in p)

    return ret


async def generate_data(
    sentence: str,
    reverse: bool = False,
    ysdd_mode: bool = True,
    normalize: bool = True,
    speed_multiple: float = 1,
    pitch_multiple: float = 1,
    pause_length: float = 0.25,
) -> SoundArrayType:
    """活字印刷"""

    sentence = sentence.lower()
    tokens_list = await parse_sentence(sentence, ysdd_mode)

    audios: List[SoundArrayType] = await asyncio.gather(
        *(
            get_pinyin_audio(token.pron, token.is_ysdd, normalize, pause_length)
            for token in tokens_list
        ),
    )
    processed_au: SoundArrayType = np.concatenate((np.array([]), *audios))

    # 音高偏移
    processed_au = modify_pitch_and_speed(processed_au, pitch_multiple, speed_multiple)

    # 倒放
    if reverse:
        processed_au = np.flip(processed_au)

    return processed_au


def save_to_bytes_io(data: SoundArrayType) -> BytesIO:
    file = BytesIO()
    sf.write(file, data, TARGET_SAMPLE_RATE, format="wav", subtype="PCM_16")
    return file


async def generate(
    sentence: str,
    reverse: bool = False,
    ysdd_mode: bool = True,
    normalize: bool = True,
    speed_multiple: float = 1,
    pitch_multiple: float = 1,
    pause_length: float = 0.25,
) -> BytesIO:
    data = await generate_data(
        sentence,
        reverse,
        ysdd_mode,
        normalize,
        speed_multiple,
        pitch_multiple,
        pause_length,
    )
    return save_to_bytes_io(data)



import json
import zipfile
from pathlib import Path
from typing import Dict, List, Optional
from zipfile import ZipFile

import httpx
from nonebot import logger

RESOURCE_DIR = Path.cwd() / "data" / "ottohzys"
if not RESOURCE_DIR.exists():
    RESOURCE_DIR.mkdir(parents=True)

TOKENS_DIR = RESOURCE_DIR / "tokens"
YSDD_TOKENS_DIR = RESOURCE_DIR / "ysddTokens"
YSDD_TOKEN_MAP_PATH = RESOURCE_DIR / "ysdd.json"

BUILTIN_RES_DIR = Path(__file__).parent / "res"
CHINGLISH_MAP_PATH = BUILTIN_RES_DIR / "chinglish.json"

RESOURCE_REPO_URL = "https://ghproxy.com/https://github.com/HanaYabuki/otto-hzys/archive/refs/heads/master.zip"


def download_resource_sync():
    cache_path = RESOURCE_DIR / "cache.zip"
    if cache_path.exists():
        cache_path.unlink()

    with cache_path.open("wb") as df:
        with httpx.stream("GET", RESOURCE_REPO_URL, follow_redirects=True) as response:
            for chunk in response.iter_bytes():
                df.write(chunk)

    def extract(
        zf: ZipFile,
        zp: Optional[zipfile.Path] = None,
        paths: Optional[List[str]] = None,
    ):
        if not zp:
            zp = zipfile.Path(zf)

        if not paths:
            paths = []
        else:
            zp = zp.joinpath(*paths)

        for p in zp.iterdir():
            real_path = RESOURCE_DIR.joinpath(*paths, p.name)
            if p.is_dir():
                real_path.mkdir(parents=True, exist_ok=True)
                extract(zf, zp, [*paths, p.name])
            else:
                real_path.write_bytes(p.read_bytes())

    with ZipFile(cache_path) as f:
        extract(f, zipfile.Path(f) / "otto-hzys-master" / "public" / "static")

    cache_path.unlink()


def check_resource_sync():
    return all(p.exists() for p in (TOKENS_DIR, YSDD_TOKENS_DIR, YSDD_TOKEN_MAP_PATH))


def check_and_download():
    if check_resource_sync():
        return

    logger.info("Resource not found, trying to download, please wait...")
    try:
        download_resource_sync()
    except Exception:
        logger.error("Failed to download resources, please download manually")
        raise

    if not check_resource_sync():
        raise RuntimeError("Failed to download resources, please download manually")

    logger.success("Successfully downloaded resources")


check_and_download()


CHINGLISH_MAP: Dict[str, str] = json.loads(
    CHINGLISH_MAP_PATH.read_text(encoding="u8"),
)
YSDD_TOKEN_MAP: Dict[str, List[str]] = json.loads(
    YSDD_TOKEN_MAP_PATH.read_text(encoding="u8"),
)



from nonebot import logger, on_shell_command
from nonebot.adapters.onebot.v11 import MessageSegment
from nonebot.exception import ParserExit
from nonebot.matcher import Matcher
from nonebot.params import ShellCommandArgs
from nonebot.rule import ArgumentParser, Namespace

from .hzys import generate

CMD_PREFIXES = ("otto", "电棍", "大电老师")
CMD_SUFFIXES = ("活字印刷", "鬼叫", "hzys")
COMMANDS = [f"{prefix}{suffix}" for prefix in CMD_PREFIXES for suffix in CMD_SUFFIXES]
CMD_PRIMARY, *CMD_ALIAS = COMMANDS


cmd_otto_parser = ArgumentParser(CMD_PRIMARY, add_help=False)
cmd_otto_parser.add_argument(
    "-r",
    "--reverse",
    action="store_true",
    help="是否倒放",
    default=False,
)
cmd_otto_parser.add_argument(
    "-Y",
    "--no-ysdd",
    action="store_true",
    help="特定语句是否 不使用 原声大碟",
    default=False,
)
cmd_otto_parser.add_argument(
    "-N",
    "--no-normalize",
    action="store_true",
    help="是否 不使用 自动音量归一化",
    default=False,
)
cmd_otto_parser.add_argument(
    "-s",
    "--speed",
    type=float,
    help="语音速度，范围 >= 0.5，默认 1.0，越小速度越慢",
    default=1.0,
)
cmd_otto_parser.add_argument(
    "-p",
    "--pitch",
    type=float,
    help="语音音调，范围 0 < pitch <= 2，默认 1.0，越大音调越高",
    default=1.0,
)
cmd_otto_parser.add_argument(
    "-P",
    "--pause",
    type=float,
    help="空格等字符的停顿时间，单位秒，默认 0.25",
    default=0.25,
)
cmd_otto_parser.add_argument("sentence", nargs="*", help="要生成语音的文本")

HELP = cmd_otto_parser.format_help()
TIP_TOO_EXTREME = "过于极端的音调和速度参数可能导致输出结果与预期不符"

cmd_otto = on_shell_command(CMD_PRIMARY, aliases=set(CMD_ALIAS), parser=cmd_otto_parser)


@cmd_otto.handle()
async def _(matcher: Matcher, err: ParserExit = ShellCommandArgs()):
    await matcher.finish(f"解析指令参数出错：{err.message}")


@cmd_otto.handle()
async def _(matcher: Matcher, args: Namespace = ShellCommandArgs()):
    sentence = " ".join(str(x) for x in args.sentence).strip()
    reverse: bool = args.reverse
    ysdd: bool = not args.no_ysdd
    normalize: bool = not args.no_normalize
    speed: float = args.speed
    pitch: float = args.pitch
    pause: float = args.pause

    if speed < 0.5:
        await matcher.finish(TIP_TOO_EXTREME)
    if not 0 < pitch <= 2:
        await matcher.finish(TIP_TOO_EXTREME)
    if not sentence:
        await matcher.finish(HELP)

    try:
        voice = await generate(
            sentence=sentence,
            reverse=reverse,
            ysdd_mode=ysdd,
            normalize=normalize,
            speed_multiple=speed,
            pitch_multiple=pitch,
            pause_length=pause,
        )
    except Exception:
        logger.exception("Failed to generate voice")
        await matcher.finish("生成语音失败")

    await matcher.finish(MessageSegment.record(voice))
